@@ -3,9 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2002-2021, University of Amsterdam
-                              VU University Amsterdam
-                              SWI-Prolog Solutions b.v.
+    Copyright (c)  2002-2023, University of Amsterdam
+			      VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,24 +35,27 @@
 */
 
 :- module(prolog_main,
-          [ main/0,
-            argv_options/3,             % +Argv, -RestArgv, -Options
-            argv_options/4,             % +Argv, -RestArgv, -Options, +ParseOpts
-            argv_usage/1,               % +Level
-            cli_parse_debug_options/2,  % +OptionsIn, -Options
-            cli_enable_development_system/0
+	  [ main/0,
+	    argv_options/3,             % +Argv, -RestArgv, -Options
+	    argv_options/4,             % +Argv, -RestArgv, -Options, +ParseOpts
+	    argv_usage/1,               % +Level
+	    cli_parse_debug_options/2,  % +OptionsIn, -Options
+            cli_debug_opt_type/3,       % -Flag, -Option, -Type
+            cli_debug_opt_help/2,       % -Option, -Message
+            cli_debug_opt_meta/2,       % -Option, -Arg
+	    cli_enable_development_system/0
           ]).
-% use autoload/1 to avoid checking these files at load time.
-:- autoload(library(debug)).
-:- autoload(library(threadutil)).
-% These are fine to be checked and loaded
-:- autoload(library(apply), [maplist/3, partition/4]).
+:- use_module(library(debug), [debug/1]).
+:- autoload(library(apply), [maplist/2, maplist/3, partition/4]).
 :- autoload(library(lists), [append/3]).
 :- autoload(library(pairs), [pairs_keys/2, pairs_values/2]).
 :- autoload(library(prolog_code), [pi_head/2]).
 :- autoload(library(prolog_debug), [spy/1]).
 :- autoload(library(dcg/high_order), [sequence//3, sequence//2]).
 :- autoload(library(option), [option/2]).
+:- if(exists_source(library(doc_markdown))).
+:- autoload(library(doc_markdown), [print_markdown/2]).
+:- endif.
 
 :- meta_predicate
     argv_options(:, -, -),
@@ -88,7 +91,7 @@ echo([H|T]) :-
 @see	library(prolog_stack) to force backtraces in case of an
 	uncaught exception.
 @see    XPCE users should have a look at library(pce_main), which
-        starts the GUI and processes events until all windows have gone.
+	starts the GUI and processes events until all windows have gone.
 */
 
 :- module_transparent
@@ -99,7 +102,22 @@ echo([H|T]) :-
 %   Call main/1 using the passed  command-line arguments. Before calling
 %   main/1  this  predicate  installs  a  signal  handler  for  =SIGINT=
 %   (Control-C) that terminates the process with status 1.
+%
+%   When main/0 is called interactively it  simply calls main/1 with the
+%   arguments. This allows for debugging scripts as follows:
+%
+%   ```
+%   $ swipl -l script.pl -- arg ...
+%   ?- gspy(suspect/1).		% setup debugging
+%   ?- main.			% run program
+%   ```
 
+main :-
+    current_prolog_flag(break_level, _),
+    !,
+    current_prolog_flag(argv, Av),
+    context_module(M),
+    M:main(Av).
 main :-
     context_module(M),
     set_signals,
@@ -152,7 +170,13 @@ interrupt(_Sig) :-
 %       or ``-``.  Type is one of:
 %
 %       - A|B
-%         Disjunctive type.
+%         Disjunctive type.  Disjunction can be used create long
+%         options with optional values.   For example, using the type
+%         ``nonneg|boolean``, for an option `http` handles ``--http``
+%         as http(true), ``--no-http`` as http(false) and ``--http=3000``
+%         as http(3000). Note that with an optional boolean a option is
+%         considered boolean unless it has a value written as
+%         ``--longopt=value``.
 %       - boolean(Default)
 %       - boolean
 %         Boolean options are special.  They do not take a value except
@@ -170,6 +194,8 @@ interrupt(_Sig) :-
 %         As `integer`.  Requires value >= 0.
 %       - natural
 %         As `integer`.  Requires value >= 1.
+%       - number
+%         Any number (integer, float, rational).
 %       - between(Low, High)
 %         If both one of Low and High is a float, convert as `float`,
 %         else convert as `integer`.  Then check the range.
@@ -183,10 +209,18 @@ interrupt(_Sig) :-
 %       - file
 %         Convert to a file name in Prolog canonical notation
 %         using prolog_to_os_filename/2.
+%       - directory
+%         Convert to a file name in Prolog canonical notation
+%         using prolog_to_os_filename/2.  No checking is done and
+%         thus this type is the same as `file`
 %       - file(Access)
 %         As `file`, and check access using access_file/2.  A value `-`
 %         is not checked for access, assuming the application handles
 %         this as standard input or output.
+%       - directory(Access)
+%         As `directory`, and check access.  Access is one of `read`
+%         `write` or `create`.  In the latter case the parent directory
+%         must exist and have write access.
 %       - term
 %         Parse option value to a Prolog term.
 %       - term(+Options)
@@ -236,9 +270,9 @@ argv_options(Argv, Positional, Options, POptions) :-
     !,
     E = error(_,_),
     catch(opt_parse(Argv, Positional, Options, POptions), E,
-          ( print_message(error, E),
-            halt(Code)
-          )).
+	  ( print_message(error, E),
+	    halt(Code)
+	  )).
 argv_options(Argv, Positional, Options, POptions) :-
     opt_parse(Argv, Positional, Options, POptions).
 
@@ -258,16 +292,16 @@ argv_untyped_options([H0|T0], R, Ops), sub_atom(H0, 0, _, _, --) =>
     Ops = [H|T],
     (   sub_atom(H0, B, _, A, =)
     ->  B2 is B-2,
-        sub_atom(H0, 2, B2, _, Name),
-        sub_string(H0, _, A,  0, Value0),
-        convert_option(Name, Value0, Value)
+	sub_atom(H0, 2, B2, _, Name),
+	sub_string(H0, _, A,  0, Value0),
+	convert_option(Name, Value0, Value)
     ;   sub_atom(H0, 2, _, 0, Name0),
-        (   sub_atom(Name0, 0, _, _, 'no-')
-        ->  sub_atom(Name0, 3, _, 0, Name),
-            Value = false
-        ;   Name = Name0,
-            Value = true
-        )
+	(   sub_atom(Name0, 0, _, _, 'no-')
+	->  sub_atom(Name0, 3, _, 0, Name),
+	    Value = false
+	;   Name = Name0,
+	    Value = true
+	)
     ),
     canonical_name(Name, PlName),
     H =.. [PlName,Value],
@@ -348,9 +382,9 @@ take_long(Long, T, Positional, Options, M, POptions) :- % --long=Value
     canonical_name(LName0, LName),
     (   in(M:opt_type(LName, Name, Type))
     ->  opt_value(Type, Long, VAtom, Value),
-        Opt =.. [Name,Value],
-        Options = [Opt|OptionsT],
-        opt_parse(T, Positional, OptionsT, M, POptions)
+	Opt =.. [Name,Value],
+	Options = [Opt|OptionsT],
+	opt_parse(T, Positional, OptionsT, M, POptions)
     ;   opt_error(unknown_option(M:LName0))
     ).
 take_long(LName0, T, Positional, Options, M, POptions) :- % --long
@@ -358,7 +392,7 @@ take_long(LName0, T, Positional, Options, M, POptions) :- % --long
     take_long_(LName, T, Positional, Options, M, POptions).
 
 take_long_(Long, T, Positional, Options, M, POptions) :- % --long
-    opt_bool_type(Long, Name, Value, M),
+    opt_bool_type(Long, Name, Value, M),                 % only boolean
     !,
     Opt =.. [Name,Value],
     Options = [Opt|OptionsT],
@@ -367,20 +401,28 @@ take_long_(Long, T, Positional, Options, M, POptions) :- % --no-long, --nolong
     (   atom_concat('no_', LName, Long)
     ;   atom_concat('no', LName, Long)
     ),
-    opt_bool_type(LName, Name, Value0, M),
+    in(M:opt_type(LName, Name, Type)),
+    type_optional_bool(Type, Value0),
     !,
     negate(Value0, Value),
     Opt =.. [Name,Value],
     Options = [Opt|OptionsT],
     opt_parse(T, Positional, OptionsT, M, POptions).
+take_long_(Long, T, Positional, Options, M, POptions) :- % --long [value]
+    in(M:opt_type(Long, Name, Type)),
+    type_optional_bool(Type, Value),
+    !,
+    Opt =.. [Name,Value],
+    Options = [Opt|OptionsT],
+    opt_parse(T, Positional, OptionsT, M, POptions).
 take_long_(Long, T, Positional, Options, M, POptions) :- % --long
     in(M:opt_type(Long, Name, Type)),
     !,
     (   T = [VAtom|T1]
     ->  opt_value(Type, Long, VAtom, Value),
-        Opt =.. [Name,Value],
-        Options = [Opt|OptionsT],
-        opt_parse(T1, Positional, OptionsT, M, POptions)
+	Opt =.. [Name,Value],
+	Options = [Opt|OptionsT],
+	opt_parse(T1, Positional, OptionsT, M, POptions)
     ;   opt_error(missing_value(Long, Type))
     ).
 take_long_(Long, _, _, _, M, _) :-
@@ -399,28 +441,39 @@ take_shorts([H|T], Argv, Positional, Options, M, POptions) :-
     !,
     (   T == []
     ->  (   Argv = [VAtom|ArgvT]
-        ->  opt_value(Type, H, VAtom, Value),
-            Opt =.. [Name,Value],
-            Options = [Opt|OptionsT],
-            take_shorts(T, ArgvT, Positional, OptionsT, M, POptions)
-        ;   opt_error(missing_value(H, Type))
-        )
+	->  opt_value(Type, H, VAtom, Value),
+	    Opt =.. [Name,Value],
+	    Options = [Opt|OptionsT],
+	    take_shorts(T, ArgvT, Positional, OptionsT, M, POptions)
+	;   opt_error(missing_value(H, Type))
+	)
     ;   atom_chars(VAtom, T),
-        opt_value(Type, H, VAtom, Value),
-        Opt =.. [Name,Value],
-        Options = [Opt|OptionsT],
-        take_shorts([], Argv, Positional, OptionsT, M, POptions)
+	opt_value(Type, H, VAtom, Value),
+	Opt =.. [Name,Value],
+	Options = [Opt|OptionsT],
+	take_shorts([], Argv, Positional, OptionsT, M, POptions)
     ).
 take_shorts([H|_], _, _, _, M, _) :-
     opt_error(unknown_option(M:H)).
 
 opt_bool_type(Opt, Name, Value, M) :-
     in(M:opt_type(Opt, Name, Type)),
+    type_bool(Type, Value).
+
+type_bool(Type, Value) :-
     (   Type == boolean
     ->  Value = true
     ;   Type = boolean(Value)
     ).
 
+type_optional_bool((A|B), Value) =>
+    (   type_optional_bool(A, Value)
+    ->  true
+    ;   type_optional_bool(B, Value)
+    ).
+type_optional_bool(Type, Value) =>
+    type_bool(Type, Value).
+
 negate(true, false).
 negate(false, true).
 
@@ -445,6 +498,8 @@ opt_convert(boolean, Spec, Value) :-
     to_bool(Spec, Value).
 opt_convert(boolean(_), Spec, Value) :-
     to_bool(Spec, Value).
+opt_convert(number, Spec, Value) :-
+    atom_number(Spec, Value).
 opt_convert(integer, Spec, Value) :-
     atom_number(Spec, Value),
     integer(Value).
@@ -464,7 +519,7 @@ opt_convert(between(Low, High), Spec, Value) :-
     (   ( float(Low) ; float(High) )
     ->  Value is float(Value0)
     ;   integer(Value0),
-        Value = Value0
+	Value = Value0
     ),
     Value >= Low, Value =< High.
 opt_convert(atom, Value, Value).
@@ -478,11 +533,16 @@ opt_convert(file(Access), Spec, Value) :-
     (   Spec == '-'
     ->  Value = '-'
     ;   prolog_to_os_filename(Value, Spec),
-        (   access_file(Value, Access)
-        ->  true
-        ;   opt_error(access_file(Spec, Access))
-        )
+	(   access_file(Value, Access)
+	->  true
+	;   opt_error(access_file(Spec, Access))
+	)
     ).
+opt_convert(directory, Spec, Value) :-
+    prolog_to_os_filename(Value, Spec).
+opt_convert(directory(Access), Spec, Value) :-
+    prolog_to_os_filename(Value, Spec),
+    access_directory(Value, Access).
 opt_convert(term, Spec, Value) :-
     term_string(Value, Spec, []).
 opt_convert(term(Options), Spec, Value) :-
@@ -492,17 +552,36 @@ opt_convert(term(Options), Spec, Value) :-
     ;   Value = Term
     ).
 
+access_directory(Dir, read) =>
+    exists_directory(Dir),
+    access_file(Dir, read).
+access_directory(Dir, write) =>
+    exists_directory(Dir),
+    access_file(Dir, write).
+access_directory(Dir, create) =>
+    (   exists_directory(Dir)
+    ->  access_file(Dir, write)
+    ;   \+ exists_file(Dir),
+        file_directory_name(Dir, Parent),
+        exists_directory(Parent),
+        access_file(Parent, write)
+    ).
+
 to_bool(true,    true).
 to_bool('True',  true).
 to_bool('TRUE',  true).
 to_bool(on,      true).
 to_bool('On',    true).
+to_bool(yes,     true).
+to_bool('Yes',   true).
 to_bool('1',     true).
 to_bool(false,   false).
 to_bool('False', false).
 to_bool('FALSE', false).
 to_bool(off,     false).
 to_bool('Off',   false).
+to_bool(no,      false).
+to_bool('No',    false).
 to_bool('0',     false).
 
 %!  argv_usage(:Level) is det.
@@ -544,6 +623,7 @@ prolog:message(opt_usage(M)) -->
 usage(M) -->
     usage_text(M:header),
     usage_line(M),
+    usage_text(M:description),
     usage_options(M),
     usage_text(M:footer).
 
@@ -556,8 +636,8 @@ usage_text(M:Which) -->
     { in(M:opt_help(help(Which), Help))
     },
     !,
-    (   {Which == header}
-    ->  user_text(M:Help), [nl]
+    (   {Which == header ; Which == description}
+    ->  user_text(M:Help), [nl, nl]
     ;   [nl], user_text(M:Help)
     ).
 usage_text(_) -->
@@ -565,7 +645,21 @@ usage_text(_) -->
 
 user_text(M:Entries) -->
     { is_list(Entries) },
+    !,
     sequence(help_elem(M), Entries).
+:- if(current_predicate(print_markdown/2)).
+user_text(_:md(Help)) -->
+    !,
+    { with_output_to(string(String),
+                     ( current_output(S),
+                       set_stream(S, tty(true)),
+                       print_markdown(Help, []))) },
+    [ '~s'-[String] ].
+:- else.
+user_text(_:md(Help)) -->
+    !,
+    [ '~w'-[Help] ].
+:- endif.
 user_text(_:Help) -->
     [ '~w'-[Help] ].
 
@@ -577,14 +671,26 @@ help_elem(_M, Elem) -->
     [ Elem ].
 
 usage_line(M) -->
+    { findall(Help, in(M:opt_help(help(usage), Help)), HelpLines)
+    },
     [ ansi(comment, 'Usage: ', []) ],
-    cmdline(M),
-    (   {in(M:opt_help(help(usage), Help))}
-    ->  user_text(M:Help)
-    ;   [ ' [options]'-[] ]
+    (   {HelpLines == []}
+    ->  cmdline(M), [ ' [options]'-[] ]
+    ;   sequence(usage_line(M), [nl], HelpLines)
     ),
     [ nl, nl ].
 
+usage_line(M, Help) -->
+    [ '~t~8|'-[] ],
+    cmdline(M),
+    user_text(M:Help).
+
+cmdline(_M) -->
+    { current_prolog_flag(app_name, App),
+      !,
+      current_prolog_flag(os_argv, [Argv0|_])
+    },
+    cmdarg(Argv0), [' '-[], ansi(bold, '~w', [App])].
 cmdline(_M) -->
     { current_prolog_flag(associated_file, AbsFile),
       file_base_name(AbsFile, Base),
@@ -620,13 +726,22 @@ usage_options(M) -->
     { findall(Opt, get_option(M, Opt), Opts),
       maplist(options_width, Opts, OptWidths),
       max_list(OptWidths, MaxOptWidth),
-      catch(tty_size(_, Width), _, Width = 80),
+      tty_width(Width),
       OptColW is min(MaxOptWidth, 30),
       HelpColW is Width-4-OptColW
     },
     [ ansi(comment, 'Options:', []), nl ],
     sequence(opt_usage(OptColW, HelpColW), [nl], Opts).
 
+% Just  catch/3  is   enough,   but    dependency   tracking   in  e.g.,
+% list_undefined/0 still considers this a missing dependency.
+:- if(current_predicate(tty_size/2)).
+tty_width(Width) :-
+     catch(tty_size(_, Width), _, Width = 80).
+:- else.
+tty_width(80).
+:- endif.
+
 opt_usage(OptColW, HelpColW, opt(_Name, Type, Short, Long, Help, Meta)) -->
     options(Type, Short, Long, Meta),
     [ '~t~*:| '-[OptColW] ],
@@ -686,8 +801,14 @@ options(Type, ShortOpt, LongOpts, Meta) -->
     sequence(option(Type, Meta), [', '-[]], Opts).
 
 option(boolean, _, Opt) -->
-    opt(Opt).
-option(_, Meta, Opt) -->
+    opt(Opt),
+    !.
+option(_Type, [Meta], Opt) -->
+    \+ { short_opt(Opt) },
+    !,
+    opt(Opt),
+    [ '[='-[], ansi(var, '~w', [Meta]), ']'-[] ].
+option(_Type, Meta, Opt) -->
     opt(Opt),
     (   { short_opt(Opt) }
     ->  [ ' '-[] ]
@@ -705,17 +826,21 @@ options_width(opt(_Name, boolean, Short, Long, _Help, _Meta), W) =>
     maplist(atom_length, Long, LLens),
     sum_list(LLens, LLen),
     W is ((SCount+LCount)-1)*2 +               % ', ' seps
-         SCount*2 +
-         LCount*2 + LLen.
+	 SCount*2 +
+	 LCount*2 + LLen.
 options_width(opt(_Name, _Type, Short, Long, _Help, Meta), W) =>
     length(Short, SCount),
     length(Long, LCount),
-    atom_length(Meta, MLen),
+    (   Meta = [MName]
+    ->  atom_length(MName, MLen0),
+        MLen is MLen0+2
+    ;   atom_length(Meta, MLen)
+    ),
     maplist(atom_length, Long, LLens),
     sum_list(LLens, LLen),
     W is ((SCount+LCount)-1)*2 +               % ', ' seps
-         SCount*3 + SCount*MLen +
-         LCount*3 + LLen + LCount*MLen.
+	 SCount*3 + SCount*MLen +
+	 LCount*3 + LLen + LCount*MLen.
 
 %!  get_option(+Module, -Opt) is multi.
 %
@@ -724,39 +849,50 @@ options_width(opt(_Name, _Type, Short, Long, _Help, Meta), W) =>
 %       opt(Name, Type, ShortFlags, Longflags, Help, Meta).
 
 get_option(M, opt(help, boolean, [h,?], [help],
-                  Help, -)) :-
+		  Help, -)) :-
     \+ in(M:opt_type(_, help, boolean)),       % user defined help
     (   in(M:opt_help(help, Help))
     ->  true
     ;   Help = "Show this help message and exit"
     ).
-get_option(M, opt(Name, Type, Short, Long, Help, Meta)) :-
+get_option(M, opt(Name, TypeName, Short, Long, Help, Meta)) :-
     findall(Name, in(M:opt_type(_, Name, _)), Names),
     list_to_set(Names, UNames),
     member(Name, UNames),
     findall(Opt-Type,
-            in(M:opt_type(Opt, Name, Type)),
-            Pairs),
+	    in(M:opt_type(Opt, Name, Type)),
+	    Pairs),
     option_type(Name, Pairs, TypeT),
-    functor(TypeT, Type, _),
+    functor(TypeT, TypeName, _),
     pairs_keys(Pairs, Opts),
     partition(short_opt, Opts, Short, Long),
     (   in(M:opt_help(Name, Help))
     ->  true
     ;   Help = ''
     ),
-    (   in(M:opt_meta(Name, Meta))
+    (   in(M:opt_meta(Name, Meta0))
     ->  true
-    ;   upcase_atom(Type, Meta)
+    ;   type_name(TypeT, Meta0)
+    ->  true
+    ;   upcase_atom(TypeName, Meta0)
+    ),
+    (   \+ type_bool(TypeT, _),
+        type_optional_bool(TypeT, _)
+    ->  Meta = [Meta0]
+    ;   Meta = Meta0
     ).
 
+type_name(oneof(Values), Name) :-
+    atomics_to_string(Values, ",", S0),
+    format(atom(Name), '{~w}', [S0]).
+
 option_type(Name, Pairs, Type) :-
     pairs_values(Pairs, Types),
     sort(Types, [Type|UTypes]),
     (   UTypes = []
     ->  true
     ;   print_message(warning,
-                      error(opt_error(multiple_types(Name, [Type|UTypes])),_))
+		      error(opt_error(multiple_types(Name, [Type|UTypes])),_))
     ).
 
 %!  in(:Goal)
@@ -888,17 +1024,80 @@ cli_parse_debug_options([H|T0], Opts) :-
 cli_parse_debug_options([H|T0], [H|T]) :-
     cli_parse_debug_options(T0, T).
 
+%!  cli_debug_opt_type(-Flag, -Option, -Type).
+%!  cli_debug_opt_help(-Option, -Message).
+%!  cli_debug_opt_meta(-Option, -Arg).
+%
+%   Implements  opt_type/3,  opt_help/2   and    opt_meta/2   for  debug
+%   arguments. Applications that wish to  use   these  features can call
+%   these predicates from their own hook.  Fot example:
+%
+%   ```
+%   opt_type(..., ..., ...).	% application types
+%   opt_type(Flag, Opt, Type) :-
+%       cli_debug_opt_type(Flag, Opt, Type).
+%   % similar for opt_help/2 and opt_meta/2
+%
+%   main(Argv) :-
+%       argv_options(Argv, Positional, Options0),
+%       cli_parse_debug_options(Options0, Options),
+%       ...
+%   ```
+
+cli_debug_opt_type(debug,       debug,       string).
+cli_debug_opt_type(spy,         spy,         string).
+cli_debug_opt_type(gspy,        gspy,        string).
+cli_debug_opt_type(interactive, interactive, boolean).
+
+cli_debug_opt_help(debug,
+                   "Call debug(Topic).  See debug/1 and debug/3. \c
+                    Multiple topics may be separated by : or ;").
+cli_debug_opt_help(spy,
+                   "Place a spy-point on Predicate. \c
+                    Multiple topics may be separated by : or ;").
+cli_debug_opt_help(gspy,
+                   "As --spy using the graphical debugger.  See tspy/1 \c
+                    Multiple topics may be separated by `;`").
+cli_debug_opt_help(interactive,
+                   "Start the Prolog toplevel after main/1 completes.").
+
+cli_debug_opt_meta(debug, 'TOPICS').
+cli_debug_opt_meta(spy,   'PREDICATES').
+cli_debug_opt_meta(gspy,  'PREDICATES').
+
+:- meta_predicate
+    spy_from_string(1, +).
+
 debug_option(interactive(true)) :-
     asserta(interactive).
-debug_option(debug(TopicS)) :-
+debug_option(debug(Spec)) :-
+    split_string(Spec, ";", "", Specs),
+    maplist(debug_from_string, Specs).
+debug_option(spy(Spec)) :-
+    split_string(Spec, ";", "", Specs),
+    maplist(spy_from_string(spy), Specs).
+debug_option(gspy(Spec)) :-
+    split_string(Spec, ";", "", Specs),
+    maplist(spy_from_string(cli_gspy), Specs).
+
+debug_from_string(TopicS) :-
     term_string(Topic, TopicS),
     debug(Topic).
-debug_option(spy(Atom)) :-
-    atom_pi(Atom, PI),
-    spy(PI).
-debug_option(gspy(Atom)) :-
-    atom_pi(Atom, PI),
-    tspy(PI).
+
+spy_from_string(Pred, Spec) :-
+    atom_pi(Spec, PI),
+    call(Pred, PI).
+
+cli_gspy(PI) :-
+    (   exists_source(library(threadutil))
+    ->  use_module(library(threadutil), [tspy/1]),
+	Goal = tspy(PI)
+    ;   exists_source(library(gui_tracer))
+    ->  use_module(library(gui_tracer), [gspy/1]),
+	Goal = gspy(PI)
+    ;   Goal = spy(PI)
+    ),
+    call(Goal).
 
 atom_pi(Atom, Module:PI) :-
     split(Atom, :, Module, PiAtom),
@@ -941,8 +1140,8 @@ cli_enable_development_system :-
     set_prolog_flag(message_ide, true),
     (   current_prolog_flag(xpce_version, _)
     ->  use_module(library(pce_dispatch)),
-        memberchk(Goal, [pce_dispatch([])]),
-        call(Goal)
+	memberchk(Goal, [pce_dispatch([])]),
+	call(Goal)
     ;   true
     ),
     set_prolog_flag(toplevel_goal, prolog).
@@ -957,7 +1156,7 @@ cli_enable_development_system :-
 
 prolog:called_by(main, [main(_)]).
 prolog:called_by(argv_options(_,_,_),
-                 [ opt_type(_,_,_),
-                   opt_help(_,_),
-                   opt_meta(_,_)
-                 ]).
+		 [ opt_type(_,_,_),
+		   opt_help(_,_),
+		   opt_meta(_,_)
+		 ]).
